"""
Took this module from Kenneth Enevoldsen for reliable and fast async completion with OpenAI's models :)

This script is adapted from OpenAI Cookbook example:
https://github.com/openai/openai-cookbook/blob/main/examples/api_request_parallel_processor.py

Using the OpenAI API to process lots of text quickly takes some care.
If you trickle in a million API requests one by one, they'll take days to complete.
If you flood a million API requests in parallel, they'll exceed the rate limits and fail with errors.
To maximize throughput, parallel requests need to be throttled to stay under rate limits.

This script parallelizes requests to the OpenAI API while throttling to stay under rate limits.

Features:
- Makes requests concurrently, to maximize throughput
- Throttles request and token usage, to stay under rate limits
- Retries failed requests up to {max_attempts} times, to avoid missing data
- Logs errors, to diagnose problems with requests

"""

import asyncio
import logging
import os
import time
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, Iterable, List, Optional

import openai
import tiktoken  # for counting tokens
from openai.error import OpenAIError, RateLimitError
from pydantic import BaseModel


def task_id_generator_function():
    """Generate integers 0, 1, 2, and so on."""
    task_id = 0
    while True:
        yield task_id
        task_id += 1


class APIResponse(BaseModel):
    """API responses, including errors, are stored in this format."""

    response: Optional[Any] = None
    errors: Optional[List[str]] = None
    task_id: Optional[int] = None

    def __bool__(self):
        return self.response is not None

    def to_json(self):
        """Convert to a json serializable object"""
        return self.dict()


@dataclass
class StatusTracker:
    """Stores metadata about the script's progress. Only one instance is created."""

    num_tasks_started: int = 0
    num_tasks_in_progress: int = 0  # script ends when this reaches 0
    num_tasks_succeeded: int = 0
    num_tasks_failed: int = 0
    num_rate_limit_errors: int = 0
    num_api_errors: int = 0  # excluding rate limit errors, counted above
    num_other_errors: int = 0
    time_of_last_rate_limit_error: float = (
        0  # used to cool off after hitting rate limits
    )


@dataclass
class APIRequest:
    """Stores an API request's input, outputs and other metadata along with a method for the API call"""

    task_id: int
    request: List[Dict[str, str]]
    token_consumption: int
    attempts_left: int
    api_errors = None

    async def call_api(
        self,
        responses: List[APIResponse],
        retry_queue: asyncio.Queue,
        status_tracker: StatusTracker,
        chatgpt_kwargs: Dict[str, Any],
    ):
        logging.info(f"Starting request {self.task_id}")
        if self.api_errors is None:
            self.api_errors = []
        try:
            response = await openai.ChatCompletion.acreate(
                messages=self.request,
                **chatgpt_kwargs,
            )
            response = response.to_dict_recursive()  # type: ignore
            status_tracker.num_tasks_succeeded += 1
            status_tracker.num_tasks_in_progress -= 1
            api_resp = APIResponse(
                response=response, task_id=self.task_id, errors=self.api_errors
            )
            logging.debug(f"Finished request {self.task_id}")
            responses.append(api_resp)
            return

        except RateLimitError as e:
            logging.warning(
                f"Request {self.task_id} failed with rate limit error {e}"
            )
            status_tracker.time_of_last_rate_limit_error = time.time()
            status_tracker.num_rate_limit_errors += 1
            self.api_errors.append(e.__repr__())

        except OpenAIError as e:
            logging.warning(f"Request {self.task_id} failed with error {e}")
            status_tracker.num_api_errors += 1
            self.api_errors.append(e.__repr__())

        if self.attempts_left:
            retry_queue.put_nowait(self)
            return
        status_tracker.num_tasks_in_progress -= 1
        status_tracker.num_tasks_failed += 1
        logging.warning(f"Request {self.task_id} failed permanently")
        responses.append(
            APIResponse(errors=self.api_errors, task_id=self.task_id)
        )


def num_tokens_consumed_from_chat_request(
    request: List[Dict[str, str]], model_name: str
) -> int:
    """Count the number of tokens in the request. Only supports completion.

    Numbers derived from
    https://github.com/openai/openai-cookbook/blob/main/examples/api_request_parallel_processor.py
    """
    encoding = tiktoken.encoding_for_model(model_name)
    # if completions request, tokens = prompt + n * max_tokens
    max_tokens = 15
    n = 1
    completion_tokens = n * max_tokens

    num_tokens = 0
    for message in request:
        num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":  # if there's a name, the role is omitted
                num_tokens -= 1  # role is always required and always 1 token
    num_tokens += 2  # every reply is primed with <im_start>assistant
    return num_tokens + completion_tokens


async def chat_parallel_request(
    requests: Iterable[List[Dict[str, str]]],
    max_requests_per_minute: int,
    max_tokens_per_minute: int,
    max_attempts_per_request: int = 5,
    **chatgpt_kwargs: Any,
):
    """Processes API requests in parallel, throttling to stay under rate limits.

    Args:
        requests: An iterable of requests, where each request is a list of messages.
        max_requests_per_minute: The maximum number of requests to send per minute.
        max_tokens_per_minute: The maximum number of tokens to send per minute.
        max_attempts_per_request: The maximum number of times to retry a request.
        chatgpt_kwargs: Keyword arguments to pass to openai.ChatCompletion.create()

    Returns:
        A coroutine that includes the APIReponse objects

    Example:
        >>> messages = [...]
        >>> output = chat_parallel_request(messages, max_requests_per_minute=30, max_tokens_per_minute=1000)
        >>> api_responses = asyncio.run(output)
    """

    logging.info(f"api arguments: {chatgpt_kwargs}")
    seconds_to_pause_after_rate_limit_error = 15
    seconds_to_sleep_each_loop = (
        0.001  # 1 ms limits max throughput to 1,000 requests per second
    )

    model_name = chatgpt_kwargs.get("model", None)
    if model_name is None:
        model_name = "gpt-3.5-turbo"
        logging.warning(f"No model specified using default model {model_name}")
        chatgpt_kwargs["model"] = model_name

    # initialize trackers
    queue_of_requests_to_retry = asyncio.Queue()
    task_id_generator = task_id_generator_function()
    status_tracker = StatusTracker()
    next_request = None  # variable to hold the next request to call

    # initialize available capacity counts
    available_request_capacity = max_requests_per_minute
    available_token_capacity = max_tokens_per_minute
    last_update_time = time.time()

    # initialize flags
    gen_not_finished = True  # Once generator is empty
    num_tokens_consumed = partial(
        num_tokens_consumed_from_chat_request, model_name=model_name
    )
    logging.debug(f"Initialization complete.")

    requests = iter(requests)
    responses = []

    while True:
        # get next request (if one is not already waiting for capacity)
        if next_request is None:
            if not queue_of_requests_to_retry.empty():
                next_request = queue_of_requests_to_retry.get_nowait()
                logging.debug(
                    f"Retrying request {next_request.task_id}: {next_request}"
                )
            elif gen_not_finished:
                try:
                    request = next(requests)
                    next_request = APIRequest(
                        task_id=next(task_id_generator),
                        request=request,
                        token_consumption=num_tokens_consumed(request),
                        attempts_left=max_attempts_per_request,
                    )
                    status_tracker.num_tasks_started += 1
                    status_tracker.num_tasks_in_progress += 1
                    logging.debug(
                        f"Reading request {next_request.task_id}: {next_request}"
                    )
                except StopIteration:
                    logging.debug("Generator is empty.")
                    gen_not_finished = False

        # update available capacity
        current_time = time.time()
        seconds_since_update = current_time - last_update_time
        available_request_capacity = min(
            available_request_capacity
            + max_requests_per_minute * seconds_since_update / 60,
            max_requests_per_minute,
        )

        available_token_capacity = min(
            available_token_capacity
            + max_tokens_per_minute * seconds_since_update / 60,
            max_tokens_per_minute,
        )
        last_update_time = current_time

        # if there is capacity, make the next request
        if next_request:
            next_request_tokens = next_request.token_consumption
            if (
                available_request_capacity >= 1
                and available_token_capacity >= next_request_tokens
            ):
                # update counters
                available_request_capacity -= 1
                available_token_capacity -= next_request_tokens
                next_request.attempts_left -= 1

                # make request
                asyncio.create_task(
                    next_request.call_api(
                        responses=responses,
                        retry_queue=queue_of_requests_to_retry,
                        status_tracker=status_tracker,
                        chatgpt_kwargs=chatgpt_kwargs,
                    )
                )
                next_request = None  # reset

        # if all requests are finished, return
        if status_tracker.num_tasks_in_progress == 0:
            break

        # main loop sleeps briefly so concurrent tasks can run
        await asyncio.sleep(seconds_to_sleep_each_loop)

        # if a rate limit error was hit recently, pause to cool down
        seconds_since_rate_limit_error = (
            time.time() - status_tracker.time_of_last_rate_limit_error
        )

        if (
            seconds_since_rate_limit_error
            < seconds_to_pause_after_rate_limit_error
        ):
            remaining_seconds_to_pause = (
                seconds_to_pause_after_rate_limit_error
                - seconds_since_rate_limit_error
            )
            await asyncio.sleep(remaining_seconds_to_pause)
            # ^e.g., if pause is 15 seconds and final limit was hit 5 seconds ago
            logging.warning(
                f"Pausing to cool down until {time.ctime(status_tracker.time_of_last_rate_limit_error + seconds_to_pause_after_rate_limit_error)}"
            )

    # after finishing, log final status
    logging.info("Parallel processing complete.")
    if status_tracker.num_tasks_failed > 0:
        logging.warning(
            f"{status_tracker.num_tasks_failed} / {status_tracker.num_tasks_started} requests failed."
        )
    if status_tracker.num_rate_limit_errors > 0:
        logging.warning(
            f"{status_tracker.num_rate_limit_errors} rate limit errors received. Consider running at a lower rate."
        )

    return responses  # type: ignore


def openai_chatcompletion(
    requests: Iterable[List[Dict[str, str]]],
    max_requests_per_minute: int = 3500,
    max_tokens_per_minute: int = 90_000,
    max_attempts_per_request: int = 5,
    chat_kwargs: Dict[str, Any] = {},
) -> List[APIResponse]:
    """
    A wrapper for the OpenAI Chat Completion API. to avoid calling asyncio.run() directly.
    """

    responses = asyncio.run(
        chat_parallel_request(
            requests=requests,
            max_requests_per_minute=max_requests_per_minute,
            max_tokens_per_minute=max_tokens_per_minute,
            max_attempts_per_request=max_attempts_per_request,
            **chat_kwargs,
        )
    )
    sorted_responses = sorted(responses, key=lambda r: r.task_id)
    return sorted_responses



"""Zero shot classification with generative language models."""
import asyncio
import warnings
from typing import Iterable

import aiohttp
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import NotFittedError
from thefuzz import process

__all__ = ["TGIZeroShotClassifier"]

default_prompt = """
### System:
You are a classification model that is really good at following
instructions and produces brief answers
that users can use as data right away.
Please follow the user's instructions as precisely as you can.
### User:
Your task will be to classify a text document into one
of the following classes: {classes}.
Please respond with a single label that you think fits
the document best.
Classify the following piece of text:
'{X}'
### Assistant:
"""


def async_run_all(requests: list[dict]):
    """Runs all requests asyncronously."""

    async def get_all(requests: list[dict]):
        async with aiohttp.ClientSession() as session:

            async def fetch(request: dict):
                async with session.post(**request) as response:
                    return await response.json()

            return await asyncio.gather(*[fetch(req) for req in requests])

    return asyncio.run(get_all(requests))


class TGIZeroShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible zero shot classification
    with HuggingFace's Text Generation Inference API.

    Parameters
    ----------
    endpoint: str, default 'http://127.0.0.1:8080/generate'
        Enpoint where the API is exposed. The default is localhost:8080.
    prompt: str, optional
        You can specify the prompt which will be used to prompt the model.
        Use placeholders to indicate where the class labels and the
        data should be placed in the prompt.
    max_new_tokens: int, default 256
        Maximum number of tokens the model should generate.
    fuzzy_match: bool, default True
        Indicates whether the output lables should be fuzzy matched
        to the learnt class labels.
        This is useful when the model isn't giving specific enough answers.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        endpoint: str = "http://127.0.0.1:8080/generate",
        prompt: str = default_prompt,
        max_new_tokens: int = 256,
        fuzzy_match: bool = True,
        device: str = "cpu",
    ):
        self.endpoint = endpoint
        self.prompt = prompt
        self.device = device
        self.classes_ = None
        self.max_new_tokens = max_new_tokens
        self.fuzzy_match = fuzzy_match

    def fit(self, X, y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.classes_ = np.array(list(set(y)))
        self.n_classes = len(self.classes_)
        return self

    def partial_fit(self, X, y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        if self.classes_ is None:
            self.classes_ = np.array(list(set(y)))
        else:
            new_labels = set(self.classes_) - set(y)
            if new_labels:
                self.classes_ = np.concatenate(self.classes_, list(new_labels))
        self.n_classes = len(self.classes_)
        return self

    def _generate_request(self, data: dict[str, str]) -> dict:
        prompt = self.prompt.format(**data)
        request_body = {
            "inputs": prompt,
            "parameters": {"max_new_tokens": self.max_new_tokens},
        }
        return dict(
            url=self.endpoint,
            headers={"Content-Type": "application/json"},
            json=request_body,
        )

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        if self.classes_ is None:
            raise NotFittedError("No class labels have been learnt yet.")
        requests = []
        classes_str = ", ".join([f"'{label}'" for label in self.classes_])
        for text in X:
            prompt_data = dict(X=text, classes=classes_str)
            requests.append(self._generate_request(prompt_data))
        responses = async_run_all(requests)
        labels = []
        for i, response in enumerate(responses):
            if "error" in response:
                error = response["error"]
                warnings.warn(
                    f"An error occurred for request {i}\n"
                    f"```{error}```\n"
                    "Assigning None."
                )
                labels.append(None)
                continue
            label = response["generated_text"].strip()
            if self.fuzzy_match and label not in self.classes_:
                label, _ = process.extractOne(label, self.classes_)  # type: ignore
            labels.append(label)
        return np.array(labels)


fewshot_prompt = """
### System:
You are a classification model that is really good at following
instructions and produces brief answers
that users can use as data right away.
Please follow the user's instructions as precisely as you can.
### User:
Your task will be to classify a text document into one
of the following classes: {classes}.
Please respond with a single label that you think fits
the document best.
Here are a couple of examples of labels assigned by experts:
{examples}
Classify the following piece of text:
'{X}'
### Assistant:
"""

example_prompt = """
Examples of texts labelled '{label}':
{examples}
"""


class TGIFewShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible few shot classification
    with HuggingFace's Text Generation Inference API.

    Parameters
    ----------
    endpoint: str, default 'http://127.0.0.1:8080/generate'
        Enpoint where the API is exposed. The default is localhost:8080.
    prompt: str, optional
        You can specify the prompt which will be used to prompt the model.
        Use placeholders to indicate where the class labels and the
        data should be placed in the prompt.
    max_new_tokens: int, default 256
        Maximum number of tokens the model should generate.
    fuzzy_match: bool, default True
        Indicates whether the output lables should be fuzzy matched
        to the learnt class labels.
        This is useful when the model isn't giving specific enough answers.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        endpoint: str = "http://127.0.0.1:8080/generate",
        prompt: str = fewshot_prompt,
        max_new_tokens: int = 256,
        fuzzy_match: bool = True,
        device: str = "cpu",
    ):
        self.endpoint = endpoint
        self.prompt = prompt
        self.device = device
        self.classes_ = None
        self.max_new_tokens = max_new_tokens
        self.fuzzy_match = fuzzy_match

    def fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.examples_ = dict()
        for text, label in zip(X, y):
            if label not in self.examples_:
                self.examples_[label] = []
            self.examples_[label].append(text)
        self.classes_ = np.array(list(self.examples_.keys()))
        self.n_classes = len(self.classes_)
        return self

    def partial_fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        for text, label in zip(X, y):
            if label not in self.examples_:
                self.examples_[label] = []
            self.examples_[label].append(text)
        self.classes_ = np.array(list(self.examples_.keys()))
        self.n_classes = len(self.classes_)
        return self

    def _generate_request(self, data: dict[str, str]) -> dict:
        prompt = self.prompt.format(**data)
        request_body = {
            "inputs": prompt,
            "parameters": {"max_new_tokens": self.max_new_tokens},
        }
        return dict(
            url=self.endpoint,
            headers={"Content-Type": "application/json"},
            json=request_body,
        )

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        if self.classes_ is None:
            raise NotFittedError("No class labels have been learnt yet.")
        requests = []
        classes_str = ", ".join([f"'{label}'" for label in self.classes_])
        text_examples = []
        for label, examples in self.examples_.items():
            examples = [f"'{example}'" for example in examples]
            subprompt = example_prompt.format(
                label=label, examples="\n".join(examples)
            )
            text_examples.append(subprompt)
        examples_subprompt = "\n".join(text_examples)
        for text in X:
            prompt_data = dict(
                X=text, classes=classes_str, examples=examples_subprompt
            )
            requests.append(self._generate_request(prompt_data))
        responses = async_run_all(requests)
        labels = []
        for i, response in enumerate(responses):
            if "error" in response:
                error = response["error"]
                warnings.warn(
                    f"An error occurred for request {i}\n"
                    f"```{error}```\n"
                    "Assigning None."
                )
                labels.append(None)
                continue
            label = response["generated_text"].strip()
            if self.fuzzy_match and label not in self.classes_:
                label, _ = process.extractOne(label, self.classes_)  # type: ignore
            labels.append(label)
        return np.array(labels)



"""Contains zero shot classification model."""
from typing import Iterable

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.exceptions import NotFittedError
from tqdm import tqdm
from transformers import pipeline

__all__ = ["ZeroShotClassifier"]


class ZeroShotClassifier(BaseEstimator, TransformerMixin, ClassifierMixin):
    """Scikit-learn compatible zero shot classification
    with HuggingFace Transformers.

    Parameters
    ----------
    model_name: str, default 'facebook/bart-large-mnli'
        Zero-shot model to load from HuggingFace.
    progress_bar: bool, default True
        Indicates whether a progress bar should be shown.
    device: str, default 'cpu'
        Indicates which device should be used for classification.
        Models are by default run on CPU.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        model_name: str = "facebook/bart-large-mnli",
        progress_bar: bool = True,
        device: str = "cpu",
    ):
        self.model_name = model_name
        self.pipe = pipeline(model=model_name, device=device)
        self.classes_ = None
        self.pandas_out = False
        self.progress_bar = progress_bar

    def fit(self, X, y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.classes_ = np.array(list(set(y)))
        self.n_classes = len(self.classes_)
        return self

    def partial_fit(self, X, y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        if self.classes_ is None:
            self.classes_ = np.array(list(set(y)))
        else:
            new_labels = set(self.classes_) - set(y)
            if new_labels:
                self.classes_ = np.concatenate(self.classes_, list(new_labels))
        self.n_classes = len(self.classes_)
        return self

    def predict_proba(self, X: Iterable[str]) -> np.ndarray:
        """Predicts class probabilities for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to predict probabilities for.

        Returns
        -------
        array of shape (n_texts, n_classes)
            Class probabilities for each text.
        """
        if self.classes_ is None:
            raise NotFittedError(
                "No class labels have been learned by the model, please fit()."
            )
        X = list(X)
        n_texts = len(X)
        res = np.empty((n_texts, self.n_classes))
        if self.progress_bar:
            X = tqdm(X)
        for i_doc, text in enumerate(X):
            out = self.pipe(text, candidate_labels=self.classes_)
            label_to_score = dict(zip(out["labels"], out["scores"]))  # type: ignore
            for i_class, label in enumerate(self.classes_):
                res[i_doc, i_class] = label_to_score[label]
        return res

    def transform(self, X: Iterable[str]):
        """Predicts class probabilities for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to predict probabilities for.

        Returns
        -------
        array of shape (n_texts, n_classes)
            Class probabilities for each text.
        """
        res = self.predict_proba(X)
        if self.pandas_out:
            import pandas as pd

            return pd.DataFrame(res, columns=self.classes_)
        else:
            return res

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        probs = self.transform(X)
        label_indices = np.argmax(probs, axis=1)
        return self.classes_[label_indices]  # type: ignore

    def get_feature_names_out(self) -> np.ndarray:
        if self.classes_ is None:
            raise NotFittedError(
                "No class labels have been learned by the model, please fit()."
            )
        return self.classes_

    @property
    def class_to_index(self) -> dict[str, int]:
        if self.classes_ is None:
            raise NotFittedError(
                "No class labels have been learned by the model, please fit()."
            )
        return dict(zip(self.classes_, range(self.n_classes)))

    def set_output(self, transform=None):
        """Set output of the transform() function to be a dataframe instead of
        a matrix if you pass transform='pandas'.
        Otherwise it will disable pandas output."""
        if transform == "pandas":
            self.pandas_out = True
        else:
            self.pandas_out = False
        return self



import os
import warnings
from typing import Iterable

import numpy as np
import openai
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import NotFittedError
from thefuzz import process

from stormtrooper.openai_async import openai_chatcompletion

default_system_prompt = """
You are a classification model that is really good at following
instructions and produces brief answers
that users can use as data right away.
Please follow the user's instructions as precisely as you can.
"""

default_prompt = """
Your task will be to classify a text document into one
of the following classes: {classes}.
Please respond with a single label that you think fits
the document best.
Classify the following piece of text:
 ```{X}```
"""

default_fewshot_prompt = """
Your task will be to classify a text document into one
of the following classes: {classes}.
Please respond with a single label that you think fits
the document best.
Here are a couple of examples of labels assigned by experts:
{examples}
Classify the following piece of text:
'{X}'
"""

example_prompt = """
Examples of texts labelled '{label}':
{examples}
"""


def parse_prompt(prompt_template: str) -> dict[str, str]:
    """Parses prompt from string template."""
    prompt_parts = prompt_template.split("###")
    prompt_parts = [part.strip() for part in prompt_parts]
    prompt_parts = [part for part in prompt_parts if part]
    prompt_mapping = {}
    for part in prompt_parts:
        role, content = part.split(":", 1)
        prompt_mapping[role] = content
    if set(prompt_mapping.keys()) != {"System", "User", "Assistant"}:
        raise ValueError(
            "Prompts should have a 'System', 'User' and 'Assistant'"
            "section in OpenAI models."
        )
    return prompt_mapping


def create_messages(prompt: dict[str, str], data: dict[str, str]):
    """Produces messages to send to the chat API
    from prompt and data to infuse."""
    messages = [
        {"role": "system", "content": prompt["System"]},
        {"role": "user", "content": prompt["User"].format(**data)},
    ]
    return messages


class OpenAIZeroShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible zero shot classification
    with OpenAI's chat language models.

    Parameters
    ----------
    model_name: str, default 'gpt-3.5-turbo'
        Name of OpenAI chat model.
    temperature: float = 1.0
        What sampling temperature to use, between 0 and 2.
        Higher values like 0.8 will make the output more random,
        while lower values like 0.2 will make it
        more focused and deterministic.
    prompt: str, optional
        You can specify the prompt which will be used to prompt the model.
        Use placeholders to indicate where the class labels and the
        data should be placed in the prompt.
    system_prompt: str, optional
        System prompt for the model.
    max_new_tokens: int, default 256
        Maximum number of tokens the model should generate.
    max_requests_per_minute: int, default 3500
        Maximum number of requests to send per minute.
    max_tokens_per_minute: int, default 90_000
        Maximum number of tokens per minute.
    max_attempts_per_request: int, default 5
        Maximum number of times a request shoulb be attempted if it fails
        for the first time.
    fuzzy_match: bool, default True
        Indicates whether the output lables should be fuzzy matched
        to the learnt class labels.
        This is useful when the model isn't giving specific enough answers.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        model_name: str = "gpt-3.5-turbo",
        temperature: float = 1.0,
        prompt: str = default_prompt,
        system_prompt: str = default_system_prompt,
        max_new_tokens: int = 256,
        max_requests_per_minute: int = 3500,
        max_tokens_per_minute: int = 90_000,
        max_attempts_per_request: int = 5,
        fuzzy_match: bool = True,
    ):
        self.model_name = model_name
        self.prompt = prompt
        self.system_prompt = system_prompt
        self.temperature = temperature
        self.classes_ = None
        self.max_new_tokens = max_new_tokens
        self.fuzzy_match = fuzzy_match
        self.max_requests_per_minute = max_requests_per_minute
        self.max_tokens_per_minute = max_tokens_per_minute
        self.max_attempts_per_request = max_attempts_per_request
        try:
            openai.api_key = os.environ["OPENAI_API_KEY"]
            openai.organization = os.environ.get("OPENAI_ORG")
        except KeyError as e:
            raise KeyError("Environment variable OPENAI_API_KEY not specified.") from e

    def fit(self, X, y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.classes_ = np.array(list(set(y)))
        self.n_classes = len(self.classes_)
        return self

    def generate_messages(self, text: str) -> list[dict[str, str]]:
        classes_in_quotes = [f"'{label}'" for label in self.classes_]
        prompt = self.prompt.format(X=text, classes=", ".join(classes_in_quotes))
        return [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": prompt},
        ]

    def partial_fit(self, X, y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: Any
            Ignored
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        if self.classes_ is None:
            self.classes_ = np.array(list(set(y)))
        else:
            new_labels = set(self.classes_) - set(y)
            if new_labels:
                self.classes_ = np.concatenate(self.classes_, list(new_labels))
        self.n_classes = len(self.classes_)
        return self

    def _produce_requests(self, X: Iterable[str]) -> Iterable[list[dict[str, str]]]:
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        for text in X:
            messages = self.generate_messages(text)
            yield messages

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        requests = list(self._produce_requests(X))
        parameters = {
            "model": self.model_name,
            "temperature": self.temperature,
            "max_tokens": self.max_new_tokens,
        }
        settings = {
            "max_requests_per_minute": self.max_requests_per_minute,
            "max_tokens_per_minute": self.max_tokens_per_minute,
            "max_attempts_per_request": self.max_attempts_per_request,
        }
        responses = openai_chatcompletion(requests, **settings, chat_kwargs=parameters)
        results = []
        for response in responses:
            if not response:
                warnings.warn(
                    f"Reponse empty due to errors: {response.api_errors}."
                    " Result will be None."
                )
                results.append(None)
                continue
            label = response.response["choices"][0]["message"]["content"].strip()
            if self.fuzzy_match and label not in self.classes_:
                label, _ = process.extractOne(label, self.classes_)
            results.append(label)
        return np.array(results)


class OpenAIFewShotClassifier(BaseEstimator, ClassifierMixin):
    """Scikit-learn compatible few shot classification
    with OpenAI's chat language models.

    Parameters
    ----------
    model_name: str, default 'gpt-3.5-turbo'
        Name of OpenAI chat model.
    temperature: float = 1.0
        What sampling temperature to use, between 0 and 2.
        Higher values like 0.8 will make the output more random,
        while lower values like 0.2 will make it
        more focused and deterministic.
    prompt: str, optional
        You can specify the prompt which will be used to prompt the model.
        Use placeholders to indicate where the class labels and the
        data should be placed in the prompt.
    system_prompt: str, optional
        System prompt for the model.
    max_new_tokens: int, default 256
        Maximum number of tokens the model should generate.
    max_requests_per_minute: int, default 3500
        Maximum number of requests to send per minute.
    max_tokens_per_minute: int, default 90_000
        Maximum number of tokens per minute.
    max_attempts_per_request: int, default 5
        Maximum number of times a request shoulb be attempted if it fails
        for the first time.
    fuzzy_match: bool, default True
        Indicates whether the output lables should be fuzzy matched
        to the learnt class labels.
        This is useful when the model isn't giving specific enough answers.

    Attributes
    ----------
    classes_: array of str
        Class names learned from the labels.
    """

    def __init__(
        self,
        model_name: str = "gpt-3.5-turbo",
        temperature: float = 1.0,
        prompt: str = default_fewshot_prompt,
        system_prompt: str = default_system_prompt,
        max_new_tokens: int = 256,
        max_requests_per_minute: int = 3500,
        max_tokens_per_minute: int = 90_000,
        max_attempts_per_request: int = 5,
        fuzzy_match: bool = True,
    ):
        self.model_name = model_name
        self.prompt = prompt
        self.temperature = temperature
        self.prompt_mapping_ = parse_prompt(prompt)
        self.classes_ = None
        self.max_new_tokens = max_new_tokens
        self.fuzzy_match = fuzzy_match
        self.max_requests_per_minute = max_requests_per_minute
        self.max_tokens_per_minute = max_tokens_per_minute
        self.max_attempts_per_request = max_attempts_per_request
        try:
            openai.api_key = os.environ["OPENAI_API_KEY"]
            openai.organization = os.environ.get("OPENAI_ORG")
        except KeyError as e:
            raise KeyError("Environment variable OPENAI_API_KEY not specified.") from e

    def fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.
            Should at least contain a representative sample
            of potential labels.

        Returns
        -------
        self
            Fitted model.
        """
        self.examples_ = dict()
        for text, label in zip(X, y):
            if label not in self.examples_:
                self.examples_[label] = []
            self.examples_[label].append(text)
        self.classes_ = np.array(list(self.examples_.keys()))
        self.n_classes = len(self.classes_)
        return self

    def generate_messages(self, text: str) -> list[dict[str, str]]:
        if self.examples_ is None:
            raise NotFittedError("No examples have been learnt yet, fit the model.")
        text_examples = []
        for label, examples in self.examples_.items():
            examples = [f"'{example}'" for example in examples]
            subprompt = example_prompt.format(label=label, examples="\n".join(examples))
            text_examples.append(subprompt)
        examples_subprompt = "\n".join(text_examples)
        classes_in_quotes = [f"'{label}'" for label in self.classes_]
        prompt = self.prompt.format(
            X=text,
            classes=", ".join(classes_in_quotes),
            examples=examples_subprompt,
        )
        return [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": prompt},
        ]

    def partial_fit(self, X: Iterable[str], y: Iterable[str]):
        """Learns class labels.
        Can learn new labels if new are encountered in the data.

        Parameters
        ----------
        X: iterable of str
            Examples to pass into the few-shot prompt.
        y: iterable of str
            Iterable of class labels.

        Returns
        -------
        self
            Fitted model.
        """
        for text, label in zip(X, y):
            if label not in self.examples_:
                self.examples_[label] = []
            self.examples_[label].append(text)
        self.classes_ = np.array(list(self.examples_.keys()))
        self.n_classes = len(self.classes_)
        return self

    def _produce_requests(self, X: Iterable[str]) -> Iterable[list[dict[str, str]]]:
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        for text in X:
            messages = self.generate_messages(text)
            yield messages

    def predict(self, X: Iterable[str]) -> np.ndarray:
        """Predicts most probable class label for given texts.

        Parameters
        ----------
        X: iterable of str
            Texts to label.

        Returns
        -------
        array of shape (n_texts)
            Array of string class labels.
        """
        if self.classes_ is None:
            raise NotFittedError(
                "Class labels have not been collected yet, please fit."
            )
        requests = list(self._produce_requests(X))
        parameters = {
            "model": self.model_name,
            "temperature": self.temperature,
            "max_tokens": self.max_new_tokens,
        }
        settings = {
            "max_requests_per_minute": self.max_requests_per_minute,
            "max_tokens_per_minute": self.max_tokens_per_minute,
            "max_attempts_per_request": self.max_attempts_per_request,
        }
        responses = openai_chatcompletion(requests, **settings, chat_kwargs=parameters)
        results = []
        for response in responses:
            if not response:
                warnings.warn(
                    f"Reponse empty due to errors: {response.api_errors}."
                    " Result will be None."
                )
                results.append(None)
                continue
            label = response.response["choices"][0]["message"]["content"].strip()
            if self.fuzzy_match and label not in self.classes_:
                label, _ = process.extractOne(label, self.classes_)
            results.append(label)
        return np.array(results)



from stormtrooper.error import NotInstalled
from stormtrooper.generative import (GenerativeFewShotClassifier,
                                     GenerativeZeroShotClassifier)
from stormtrooper.text2text import (Text2TextFewShotClassifier,
                                    Text2TextZeroShotClassifier)
from stormtrooper.text_gen_inference import (TGIFewShotClassifier,
                                             TGIZeroShotClassifier)
from stormtrooper.zero_shot import ZeroShotClassifier

try:
    from stormtrooper.set_fit import (SetFitFewShotClassifier,
                                      SetFitZeroShotClassifier)
except ModuleNotFoundError:
    SetFitZeroShotClassifier = NotInstalled(
        "SetFitZeroShotClassifier", "setfit"
    )
    SetFitFewShotClassifier = NotInstalled("SetFitFewShotClassifier", "setfit")

try:
    from stormtrooper.openai import (OpenAIFewShotClassifier,
                                     OpenAIZeroShotClassifier)
except ModuleNotFoundError:
    OpenAIZeroShotClassifier = NotInstalled(
        "OpenAIZeroShotClassifier", "openai"
    )
    OpenAIFewShotClassifier = NotInstalled("OpenAIFewShotClassifier", "openai")

__all__ = [
    "GenerativeZeroShotClassifier",
    "GenerativeFewShotClassifier",
    "Text2TextZeroShotClassifier",
    "Text2TextFewShotClassifier",
    "ZeroShotClassifier",
    "SetFitZeroShotClassifier",
    "SetFitFewShotClassifier",
    "OpenAIFewShotClassifier",
    "OpenAIZeroShotClassifier",
    "TGIZeroShotClassifier",
    "TGIFewShotClassifier",
]
